/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * This file is a part of the libzmf project.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include "ZMF4Parser.h"
#include <algorithm>

#define NO_ID 0xffffffff

namespace
{

template<typename T>
double um2in(T micrometers)
{
  return micrometers / 1000 / 25.4;
}

}

namespace libzmf
{

ZMF4Parser::ZMF4Parser(const RVNGInputStreamPtr &input, ZMFCollector *collector)
  : m_input(input)
  , m_collector(collector)
  , m_header()
  , m_pageSettings()
  , m_currentObjectHeader()
  , m_pens()
  , m_fills()
  , m_transparencies()
  , m_shadows()
{
}

bool ZMF4Parser::parse()
{
  if (!m_header.load(m_input) || !m_header.isSupported())
    return false;

  m_collector->startDocument();

  if (m_header.startBitmapOffset() > 0)
  {
    seek(m_input, m_header.startBitmapOffset());

    readBitmap();
  }
  else
  {
    seek(m_input, m_header.startContentOffset());
  }

  readDocumentSettings();

  while (!m_input->isEnd())
  {
    readPage();
  }

  m_collector->endDocument();

  return true;
}

ZMF4Parser::ObjectType ZMF4Parser::parseObjectType(uint8_t type)
{
  switch (type)
  {
  case 0xa:
    return ObjectType::FILL;
  case 0xb:
    return ObjectType::TRANSPARENCY;
  case 0xc:
    return ObjectType::PEN;
  case 0xd:
    return ObjectType::SHADOW;
  case 0x21:
    return ObjectType::PAGE_START;
  case 0x22:
    return ObjectType::MASTER_PAGE;
  case 0x23:
    return ObjectType::PAGE_END;
  case 0x24:
    return ObjectType::LAYER_START;
  case 0x25:
    return ObjectType::LAYER_END;
  case 0x27:
    return ObjectType::DOCUMENT_SETTINGS;
  case 0x28:
    return ObjectType::COLOR_PALETTE;
  case 0x32:
    return ObjectType::RECTANGLE;
  case 0x33:
    return ObjectType::ELLIPSE;
  case 0x36:
    return ObjectType::CURVE;
  case 0x41:
    return ObjectType::GROUP_START;
  case 0x42:
    return ObjectType::GROUP_END;
  default:
    ZMF_DEBUG_MSG(("Unknown object type 0x%x\n", type));
    return ObjectType::UNKNOWN;
  }
}

ZMF4Parser::ObjectHeader ZMF4Parser::readObjectHeader()
{
  ObjectHeader header;

  header.size = readU32(m_input);
  header.type = parseObjectType(readU8(m_input));

  skip(m_input, 7);

  header.refObjCount = readU32(m_input);

  skip(m_input, 8);

  uint32_t id = readU32(m_input);
  if (id != NO_ID)
  {
    header.id = id;
  }

  header.nextObjectOffset = m_input->tell() + (header.size - 28);

  return header;
}

std::vector<ZMF4Parser::ObjectRef> ZMF4Parser::readObjectRefs()
{
  std::vector<ZMF4Parser::ObjectRef> refs(m_currentObjectHeader.refObjCount);

  if (m_currentObjectHeader.refObjCount > 0)
  {
    seek(m_input, m_currentObjectHeader.nextObjectOffset - 8 * m_currentObjectHeader.refObjCount);

    for (uint32_t i = 0; i < m_currentObjectHeader.refObjCount; i++)
    {
      refs[i].id = readU32(m_input);
    }

    for (uint32_t i = 0; i < m_currentObjectHeader.refObjCount; i++)
    {
      refs[i].tag = readU32(m_input);
    }

    refs.erase(std::remove_if(refs.begin(), refs.end(),
                              [](const ObjectRef &ref)
    {
      return ref.id == NO_ID;
    }), refs.end());
  }

  return refs;
}

Style ZMF4Parser::readStyle()
{
  auto refs = readObjectRefs();

  Style style;

  for (const auto &ref : refs)
  {
    switch (ref.tag)
    {
    case 1:
      if (m_fills.find(ref.id) != m_fills.end())
      {
        style.fill = m_fills[ref.id];
      }
      else
      {
        ZMF_DEBUG_MSG(("Fill with ID 0x%x not found\n", ref.id));
      }
      break;
    case 2:
      if (m_pens.find(ref.id) != m_pens.end())
      {
        style.pen = m_pens[ref.id];
      }
      else
      {
        ZMF_DEBUG_MSG(("Pen with ID 0x%x not found\n", ref.id));
      }
      break;
    case 3:
      if (m_shadows.find(ref.id) != m_shadows.end())
      {
        style.shadow = m_shadows[ref.id];
      }
      else
      {
        ZMF_DEBUG_MSG(("Shadow with ID 0x%x not found\n", ref.id));
      }
      break;
    case 4:
      if (m_transparencies.find(ref.id) != m_transparencies.end())
      {
        style.transparency = m_transparencies[ref.id];
      }
      else
      {
        ZMF_DEBUG_MSG(("Transparency with ID 0x%x not found\n", ref.id));
      }
      break;
    default:
      ZMF_DEBUG_MSG(("unknown style ref tag %u\n", ref.tag));
      break;
    }
  }

  return style;
}

Point ZMF4Parser::readPoint()
{
  double x = um2in(readS32(m_input));
  double y = um2in(readS32(m_input));
  return Point(x, y);
}

std::vector<Point> ZMF4Parser::readBoundingBox()
{
  std::vector<Point> points;

  for (uint32_t i = 0; i < 4; i++)
  {
    points.push_back(readPoint());
  }

  return points;
}

Point ZMF4Parser::readBoundingBoxTopLeftPoint()
{
  std::vector<Point> points = readBoundingBox();

  std::nth_element(points.begin(), points.begin(), points.end(),
                   [](const Point &a, const Point &b)
  {
    return a.x + a.y < b.x + b.y;
  });

  return points[0];
}

Color ZMF4Parser::readColor()
{
  Color color;
  color.red = readU8(m_input);
  color.green = readU8(m_input);
  color.blue = readU8(m_input);
  return color;
}

Gradient ZMF4Parser::readGradient(uint32_t type)
{
  Gradient gradient;

  switch (type)
  {
  case 2:
    gradient.type = GradientType::LINEAR;
    break;
  case 3:
    gradient.type = GradientType::RADIAL;
    break;
  case 4:
    gradient.type = GradientType::CONICAL;
    break;
  case 5:
    gradient.type = GradientType::CROSS;
    break;
  case 6:
    gradient.type = GradientType::RECTANGULAR;
    break;
  case 7:
    gradient.type = GradientType::FLEXIBLE;
    break;
  default:
    ZMF_DEBUG_MSG(("unknown gradient type %u\n", type));
    gradient.type = GradientType::LINEAR;
    break;
  }

  skip(m_input, 4);

  uint32_t stopCount = readU32(m_input);

  skip(m_input, 4);

  gradient.center.x = readFloat(m_input);
  gradient.center.y = readFloat(m_input);

  gradient.angle = readFloat(m_input);

  skip(m_input, 4);

  gradient.stops.reserve(stopCount);
  for (uint32_t i = 0; i < stopCount; i++)
  {
    GradientStop stop;

    skip(m_input, 4);
    stop.color = readColor();

    skip(m_input, 5);
    stop.offset = readFloat(m_input);

    gradient.stops.push_back(stop);
  }

  return gradient;
}

void ZMF4Parser::readBitmap()
{
  skip(m_input, 2);

  uint32_t size = readU32(m_input);

  skip(m_input, size - 2 - 4);
}

void ZMF4Parser::readDocumentSettings()
{
  ObjectHeader header = readObjectHeader();

  if (header.type != ObjectType::DOCUMENT_SETTINGS)
    throw GenericException();

  skip(m_input, 32);

  Color color = readColor();

  skip(m_input, 5);

  double pageWidth = um2in(readU32(m_input));
  double pageHeight = um2in(readU32(m_input));

  skip(m_input, 68);

  double leftOffset = um2in(readU32(m_input));
  double topOffset = um2in(readU32(m_input));

  m_pageSettings = ZMFPageSettings(pageWidth, pageHeight, leftOffset, topOffset, color);

  seek(m_input, header.nextObjectOffset);
}

void ZMF4Parser::readPage()
{
  ObjectHeader startPageHeader;

  // skip color palettes, all used colors are included in Fill/Pen/...
  while (true)
  {
    startPageHeader = readObjectHeader();
    if (startPageHeader.type == ObjectType::COLOR_PALETTE)
    {
      seek(m_input, startPageHeader.nextObjectOffset);
    }
    else
    {
      break;
    }
  }

  if (startPageHeader.type != ObjectType::PAGE_START)
    throw GenericException();

  m_collector->startPage(m_pageSettings);

  seek(m_input, startPageHeader.nextObjectOffset);

  while (true)
  {
    ObjectHeader header = readObjectHeader();

    switch (header.type)
    {
    case ObjectType::MASTER_PAGE:
      seek(m_input, header.nextObjectOffset);
      break;
    case ObjectType::PAGE_END:
      m_collector->endPage();
      if (!m_input->isEnd())
        seek(m_input, header.nextObjectOffset);
      return;
    case ObjectType::LAYER_START:
      readLayer(header);
      break;
    default:
      throw GenericException();
    }
  }
}

void ZMF4Parser::readLayer(const ZMF4Parser::ObjectHeader &layerStartObjHeader)
{
  if (layerStartObjHeader.type != ObjectType::LAYER_START)
    throw GenericException();

  m_collector->startLayer();

  seek(m_input, layerStartObjHeader.nextObjectOffset);

  while (true)
  {
    m_currentObjectHeader = readObjectHeader();

    switch (m_currentObjectHeader.type)
    {
    case ObjectType::LAYER_END:
      seek(m_input, m_currentObjectHeader.nextObjectOffset);
      m_collector->endLayer();
      return;
    case ObjectType::FILL:
      readFill();
      break;
    case ObjectType::TRANSPARENCY:
      readTransparency();
      break;
    case ObjectType::PEN:
      readPen();
      break;
    case ObjectType::SHADOW:
      readShadow();
      break;
    case ObjectType::RECTANGLE:
      readRectangle();
      break;
    case ObjectType::ELLIPSE:
      readEllipse();
      break;
    case ObjectType::CURVE:
      readCurve();
      break;
    case ObjectType::GROUP_START:
      m_collector->startGroup();
      break;
    case ObjectType::GROUP_END:
      m_collector->endGroup();
      break;
    default:
      break;
    }

    seek(m_input, m_currentObjectHeader.nextObjectOffset);
  }
}

void ZMF4Parser::readFill()
{
  if (!m_currentObjectHeader.id)
  {
    ZMF_DEBUG_MSG(("Fill without ID, offset %ld\n", m_input->tell()));
    return;
  }

  skip(m_input, 8);

  uint32_t type = readU32(m_input);

  if (type == 1)
  {
    skip(m_input, 8);

    Color color = readColor();

    m_fills[m_currentObjectHeader.id.value()] = color;
  }
  else if (type >= 2 && type <= 7)
  {
    Gradient gradient = readGradient(type);

    m_fills[m_currentObjectHeader.id.value()] = gradient;
  }
  else
  {
    ZMF_DEBUG_MSG(("unknown fill type %u\n", type));
  }
}

void ZMF4Parser::readTransparency()
{
  if (!m_currentObjectHeader.id)
  {
    ZMF_DEBUG_MSG(("Transparency without ID, offset %ld\n", m_input->tell()));
    return;
  }

  skip(m_input, 8);

  uint32_t type = readU32(m_input);

  if (type == 1)
  {
    skip(m_input, 8);

    Transparency transparency;
    transparency.color = readColor();

    m_transparencies[m_currentObjectHeader.id.value()] = transparency;
  }
}

void ZMF4Parser::readPen()
{
  if (!m_currentObjectHeader.id)
  {
    ZMF_DEBUG_MSG(("Pen without ID, offset %ld\n", m_input->tell()));
    return;
  }

  Pen pen;

  skip(m_input, 12);

  uint32_t lineJoin = readU32(m_input);

  switch (lineJoin)
  {
  default:
    ZMF_DEBUG_MSG(("Unknown line join type %u\n", lineJoin));
  case 0:
    pen.lineJoinType = LineJoinType::MITER;
    break;
  case 1:
    pen.lineJoinType = LineJoinType::ROUND;
    break;
  case 2:
    pen.lineJoinType = LineJoinType::BEVEL;
    break;
  }

  uint32_t lineCap = readU32(m_input);

  switch (lineCap)
  {
  default:
    ZMF_DEBUG_MSG(("Unknown line cap type %u\n", lineCap));
  case 0:
    pen.lineCapType = LineCapType::BUTT;
    break;
  case 1:
    pen.lineCapType = LineCapType::FLAT;
    break;
  case 2:
    pen.lineCapType = LineCapType::ROUND;
    break;
  case 3:
    pen.lineCapType = LineCapType::POINTED;
    break;
  }

  skip(m_input, 4);

  pen.width = um2in(readU32(m_input));

  skip(m_input, 4);

  pen.color = readColor();

  m_pens[m_currentObjectHeader.id.value()] = pen;
}

void ZMF4Parser::readShadow()
{
  if (!m_currentObjectHeader.id)
  {
    ZMF_DEBUG_MSG(("Shadow without ID, offset %ld\n", m_input->tell()));
    return;
  }

  Shadow shadow;

  skip(m_input, 8);

  uint32_t type = readU32(m_input);

  shadow.offset = readPoint();
  shadow.angle = readFloat(m_input);

  switch (type)
  {
  case 1: // Color
  case 3: // Soft
  {
    skip(m_input, 4);

    shadow.color = readColor();

    if (type == 1)
      break;

    skip(m_input, 5);

    shadow.opacity = 1.0 - readFloat(m_input);

    // todo: blur
  }
  break;
  case 2: // Brightness
  case 4: // Transparent
  {
    shadow.opacity = 1.0 - readFloat(m_input);

    // todo: ...
  }
  break;
  default:
    ZMF_DEBUG_MSG(("Unknown shadow type %u\n", type));
    break;
  }

  m_shadows[m_currentObjectHeader.id.value()] = shadow;
}

void ZMF4Parser::readCurve()
{
  skip(m_input, 52);

  uint32_t componentsCount = readU32(m_input);

  std::vector<Curve> curves(componentsCount);

  uint32_t pointsCount = 0;

  for (uint32_t i = 0; i < componentsCount; i++)
  {
    skip(m_input, 8);

    uint32_t componentPointsCount = readU32(m_input);

    pointsCount += componentPointsCount;

    curves[i].points.resize(componentPointsCount);
    curves[i].sectionTypes.reserve(componentPointsCount - 1);

    curves[i].closed = static_cast<bool>(readU32(m_input));
  }

  for (auto &curve : curves)
  {
    for (size_t i = 0; i < curve.points.size(); i++)
    {
      curve.points[i] = readPoint();
    }
  }

  for (auto &curve : curves)
  {
    while (true)
    {
      uint32_t type = readU32(m_input);
      if (type == 0x64)
      {
        break;
      }
      else
      {
        switch (type)
        {
        default:
          ZMF_DEBUG_MSG(("Unknown point type %u\n", type));
        case 1:
          curve.sectionTypes.push_back(CurveType::LINE);
          break;
        case 2:
          curve.sectionTypes.push_back(CurveType::BEZIER_CURVE);
          skip(m_input, 8);
          break;
        }
      }
    }
  }

  m_collector->setStyle(readStyle());

  m_collector->collectPath(curves);
}

void ZMF4Parser::readRectangle()
{
  skip(m_input, 8);

  Curve curve;
  curve.points = readBoundingBox();
  curve.sectionTypes = std::vector<CurveType>(curve.points.size() - 1, CurveType::LINE);
  curve.closed = true;

  m_collector->setStyle(readStyle());

  m_collector->collectPath(curve);
}

void ZMF4Parser::readEllipse()
{
  double width = um2in(readU32(m_input));
  double height = um2in(readU32(m_input));

  Point topLeft = readBoundingBoxTopLeftPoint();

  float beginAngle = readFloat(m_input);
  float endAngle = readFloat(m_input);

  bool closed = !readU8(m_input);

  double rx = width / 2;
  double ry = height / 2;

  Point c(topLeft.x + rx, topLeft.y + ry);

  m_collector->setStyle(readStyle());

  if (ZMF_ALMOST_ZERO(beginAngle) && ZMF_ALMOST_ZERO(endAngle))
  {
    m_collector->collectEllipse(c, rx, ry);
  }
  else
  {
    m_collector->collectArc(c, rx, ry, beginAngle, endAngle, closed);
  }
}

}

/* vim:set shiftwidth=2 softtabstop=2 expandtab: */
