/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * This file is a part of the libzmf project.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#ifndef ZMF4PARSER_H_INCLUDED
#define ZMF4PARSER_H_INCLUDED

#include <librevenge/librevenge.h>
#include "libzmf_utils.h"
#include "ZMF4Header.h"
#include "ZMFCollector.h"
#include "ZMFTypes.h"
#include <vector>
#include <map>
#include <boost/optional.hpp>

namespace libzmf
{

class ZMF4Parser
{
  // disable copying
  ZMF4Parser(const ZMF4Parser &other);
  ZMF4Parser &operator=(const ZMF4Parser &other);

public:
  ZMF4Parser(const RVNGInputStreamPtr &input, ZMFCollector *collector);
  bool parse();

private:
  enum class ObjectType
  {
    UNKNOWN,
    FILL,
    TRANSPARENCY,
    PEN,
    SHADOW,
    PAGE_START,
    MASTER_PAGE,
    PAGE_END,
    LAYER_START,
    LAYER_END,
    DOCUMENT_SETTINGS,
    COLOR_PALETTE,
    RECTANGLE,
    ELLIPSE,
    CURVE,
    GROUP_START,
    GROUP_END
  };

  struct ObjectHeader
  {
    ObjectType type;
    uint32_t size;
    uint32_t nextObjectOffset;
    boost::optional<uint32_t> id;
    uint32_t refObjCount;

    ObjectHeader()
      : type()
      , size(0)
      , nextObjectOffset(0)
      , id()
      , refObjCount(0)
    { }
  };

  struct ObjectRef
  {
    uint32_t id;
    uint32_t tag;
  };

  static ObjectType parseObjectType(uint8_t type);

  ObjectHeader readObjectHeader();

  std::vector<ObjectRef> readObjectRefs();

  Style readStyle();

  Point readPoint();

  std::vector<Point> readBoundingBox();
  Point readBoundingBoxTopLeftPoint();

  Color readColor();
  Gradient readGradient(uint32_t type);

  void readBitmap();

  void readDocumentSettings();
  void readPage();
  void readLayer(const ObjectHeader &layerStartObjHeader);

  void readPen();
  void readFill();
  void readTransparency();
  void readShadow();

  void readCurve();
  void readRectangle();
  void readEllipse();

  const RVNGInputStreamPtr m_input;
  ZMFCollector *m_collector;

  ZMF4Header m_header;

  ZMFPageSettings m_pageSettings;

  ObjectHeader m_currentObjectHeader;

  std::map<uint32_t, Pen> m_pens;
  std::map<uint32_t, Fill> m_fills;
  std::map<uint32_t, Transparency> m_transparencies;
  std::map<uint32_t, Shadow> m_shadows;
};

}

#endif // ZMF4PARSER_H_INCLUDED

/* vim:set shiftwidth=2 softtabstop=2 expandtab: */
