/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * This file is a part of the libzmf project.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include "ZMFCollector.h"

#include <boost/math/constants/constants.hpp>
#include <boost/variant.hpp>

namespace libzmf
{

namespace
{

Point calculateEllipsePoint(const Point &c, double rx, double ry, double angle)
{
  return Point(c.x + rx * std::cos(angle), c.y + ry * std::sin(angle));
}


double rad2deg(double value)
{
  using namespace boost::math::double_constants;

  while (2 * pi <= value)
    value -= 2 * pi;
  while (0 > value)
    value += 2 * pi;

  return value / pi * 180;
}

double distance(const Point &p1, const Point &p2)
{
  return std::hypot(p2.x - p1.x, p2.y - p1.y);
}

class FillWriter : public boost::static_visitor<void>
{
public:
  explicit FillWriter(librevenge::RVNGPropertyList &propList,
                      const boost::optional<Transparency> &transparency)
    : m_propList(propList)
    , m_transparency(transparency)
  { }

  void operator()(const Color &color)
  {
    m_propList.insert("draw:fill", "solid");
    m_propList.insert("draw:fill-color", color.toString());

    if (m_transparency)
    {
      m_propList.insert("draw:opacity", m_transparency.value().opacity(), librevenge::RVNG_PERCENT);
    }
  }

  void operator()(const Gradient &gradient)
  {
    using namespace boost::math::double_constants;

    if (gradient.stops.size() < 2)
      return;

    m_propList.insert("draw:fill", "gradient");

    auto stops = gradient.stops;
    std::sort(stops.begin(), stops.end(),
              [&gradient](const GradientStop &s1, const GradientStop &s2)
    {
      return gradient.type == GradientType::LINEAR ? s1.offset < s2.offset : s1.offset > s2.offset;
    });
    if (gradient.type != GradientType::LINEAR)
    {
      for (auto &stop : stops)
      {
        stop.offset = 1.0 - stop.offset;
      }
    }

    librevenge::RVNGPropertyListVector gradientVector;
    for (const auto &stop : stops)
    {
      librevenge::RVNGPropertyList grad;
      grad.insert("svg:offset", stop.offset, librevenge::RVNG_PERCENT);
      grad.insert("svg:stop-color", stop.color.toString());
      grad.insert("svg:stop-opacity", m_transparency ? m_transparency.value().opacity() : 1.0, librevenge::RVNG_PERCENT);
      gradientVector.append(grad);
    }

    switch (gradient.type)
    {
    default:
    case GradientType::LINEAR:
      m_propList.insert("draw:style", "linear");
      m_propList.insert("draw:angle", rad2deg(gradient.angle + pi / 2.0));
      m_propList.insert("svg:linearGradient", gradientVector);
      break;
    case GradientType::RADIAL:
      m_propList.insert("draw:style", "radial");
      m_propList.insert("draw:cx", gradient.center.x, librevenge::RVNG_PERCENT);
      m_propList.insert("draw:cy", gradient.center.y, librevenge::RVNG_PERCENT);
      m_propList.insert("draw:border", 0.25 - distance(gradient.center, Point(0.5, 0.5)), librevenge::RVNG_PERCENT);
      m_propList.insert("svg:radialGradient", gradientVector);
      break;
    }
  }

private:
  librevenge::RVNGPropertyList &m_propList;

  const boost::optional<Transparency> &m_transparency;
};

}

ZMFCollector::ZMFCollector(librevenge::RVNGDrawingInterface *painter)
  : m_painter(painter)
  , m_pageSettings()
  , m_isDocumentStarted(false)
  , m_isPageStarted(false)
  , m_isLayerStarted(false)
  , m_style()
{
}

ZMFCollector::~ZMFCollector()
{
  if (m_isDocumentStarted)
  {
    endDocument();
  }
}

void ZMFCollector::startDocument()
{
  if (m_isDocumentStarted)
    return;

  librevenge::RVNGPropertyList propList;

  m_painter->startDocument(propList);

  m_isDocumentStarted = true;
}

void ZMFCollector::endDocument()
{
  if (!m_isDocumentStarted)
    return;

  if (m_isPageStarted)
    endPage();

  m_painter->endDocument();

  m_isDocumentStarted = false;
}

void ZMFCollector::startPage(const ZMFPageSettings &pageSettings)
{
  if (m_isPageStarted)
    return;

  if (m_isLayerStarted)
    endLayer();

  librevenge::RVNGPropertyList propList;

  propList.insert("svg:width", pageSettings.width);
  propList.insert("svg:height", pageSettings.height);

  propList.insert("draw:fill", "solid");
  propList.insert("draw:fill-color", pageSettings.color.toString());

  m_painter->startPage(propList);

  m_pageSettings = pageSettings;

  m_isPageStarted = true;
}

void ZMFCollector::endPage()
{
  if (!m_isPageStarted)
    return;

  m_painter->endPage();

  m_isPageStarted = false;
}

void ZMFCollector::startLayer()
{
  if (m_isLayerStarted)
    return;

  librevenge::RVNGPropertyList propList;

  m_painter->startLayer(propList);

  m_isLayerStarted = true;
}

void ZMFCollector::endLayer()
{
  if (!m_isLayerStarted)
    return;

  m_painter->endLayer();

  m_isLayerStarted = false;
}

void ZMFCollector::startGroup()
{
  librevenge::RVNGPropertyList propList;

  m_painter->openGroup(propList);
}

void ZMFCollector::endGroup()
{
  m_painter->closeGroup();
}

void ZMFCollector::setStyle(const Style &style)
{
  m_style = style;
}

void ZMFCollector::collectPath(const std::vector<Curve> &curves)
{
  librevenge::RVNGPropertyListVector path;

  for (const auto &curve : curves)
  {
    if (curve.points.size() < 2)
      continue;

    {
      librevenge::RVNGPropertyList pathPart;
      pathPart.insert("librevenge:path-action", "M");
      pathPart.insert("svg:x", pageX(curve.points[0].x));
      pathPart.insert("svg:y", pageY(curve.points[0].y));

      path.append(pathPart);
    }

    size_t i = 1;

    for (const auto &sectionType : curve.sectionTypes)
    {
      librevenge::RVNGPropertyList pathPart;

      switch (sectionType)
      {
      default:
      case CurveType::LINE:
        if (i >= curve.points.size())
        {
          ZMF_DEBUG_MSG(("Unexpected end of curve points\n"));
          break;
        }
        pathPart.insert("librevenge:path-action", "L");
        pathPart.insert("svg:x", pageX(curve.points[i].x));
        pathPart.insert("svg:y", pageY(curve.points[i].y));
        i++;
        break;
      case CurveType::BEZIER_CURVE:
        if (i + 2 >= curve.points.size())
        {
          ZMF_DEBUG_MSG(("Unexpected end of curve points\n"));
          break;
        }
        pathPart.insert("librevenge:path-action", "C");
        pathPart.insert("svg:x1", pageX(curve.points[i].x));
        pathPart.insert("svg:y1", pageY(curve.points[i].y));
        pathPart.insert("svg:x2", pageX(curve.points[i + 1].x));
        pathPart.insert("svg:y2", pageY(curve.points[i + 1].y));
        pathPart.insert("svg:x", pageX(curve.points[i + 2].x));
        pathPart.insert("svg:y", pageY(curve.points[i + 2].y));
        i += 3;
        break;
      }

      path.append(pathPart);
    }

    if (curve.closed)
    {
      librevenge::RVNGPropertyList pathPart;
      pathPart.insert("librevenge:path-action", "Z");

      path.append(pathPart);
    }
  }

  librevenge::RVNGPropertyList pathPropList;

  pathPropList.insert("svg:d", path);

  writeStyle(pathPropList, m_style,
             !std::any_of(curves.begin(), curves.end(), [](const Curve &c)
  {
    return c.closed;
  }));

  m_painter->setStyle(pathPropList);
  m_painter->drawPath(pathPropList);
}

void ZMFCollector::collectPath(const Curve &curve)
{
  collectPath(std::vector<Curve> { curve });
}

void ZMFCollector::collectEllipse(const Point &c, double rx, double ry)
{
  librevenge::RVNGPropertyList ellipse;

  ellipse.insert("svg:cx", pageX(c.x));
  ellipse.insert("svg:cy", pageY(c.y));
  ellipse.insert("svg:rx", rx);
  ellipse.insert("svg:ry", ry);

  writeStyle(ellipse, m_style);

  m_painter->setStyle(ellipse);
  m_painter->drawEllipse(ellipse);
}

void ZMFCollector::collectArc(const Point &c, double rx, double ry, double beginAngle, double endAngle, bool closed)
{
  using namespace boost::math::double_constants;

  auto beginPoint = calculateEllipsePoint(c, rx, ry, beginAngle);
  auto endPoint = calculateEllipsePoint(c, rx, ry, endAngle);

  double angleDiff = std::fabs(endAngle - beginAngle);
  bool largeArc = (beginAngle < endAngle && angleDiff > pi) ||
                  (beginAngle > endAngle && angleDiff < pi);

  librevenge::RVNGPropertyListVector path;

  {
    librevenge::RVNGPropertyList pathStart;
    pathStart.insert("librevenge:path-action", "M");
    pathStart.insert("svg:x", pageX(beginPoint.x));
    pathStart.insert("svg:y", pageY(beginPoint.y));

    path.append(pathStart);
  }

  {
    librevenge::RVNGPropertyList arc;
    arc.insert("librevenge:path-action", "A");
    arc.insert("svg:rx", rx);
    arc.insert("svg:ry", ry);
    arc.insert("librevenge:large-arc", largeArc ? 1 : 0);
    arc.insert("librevenge:sweep", 1);
    arc.insert("svg:x", pageX(endPoint.x));
    arc.insert("svg:y", pageY(endPoint.y));

    path.append(arc);
  }

  if (closed)
  {
    librevenge::RVNGPropertyList pathEndLine;
    pathEndLine.insert("librevenge:path-action", "L");
    pathEndLine.insert("svg:x", pageX(c.x));
    pathEndLine.insert("svg:y", pageY(c.y));

    path.append(pathEndLine);

    librevenge::RVNGPropertyList pathEnd;
    pathEnd.insert("librevenge:path-action", "Z");

    path.append(pathEnd);
  }

  librevenge::RVNGPropertyList pathPropList;

  pathPropList.insert("svg:d", path);

  writeStyle(pathPropList, m_style, !closed);

  m_painter->setStyle(pathPropList);
  m_painter->drawPath(pathPropList);
}

double ZMFCollector::pageX(double canvasX)
{
  return canvasX - m_pageSettings.leftOffset;
}

double ZMFCollector::pageY(double canvasY)
{
  return canvasY - m_pageSettings.topOffset;
}

// noFill is used to ignore Fill object even if it exists
// some implementations (such as svg in web browsers) try to fill not closed paths if fill is specified
void ZMFCollector::writeStyle(librevenge::RVNGPropertyList &propList, const Style &style, bool noFill)
{
  propList.insert("draw:stroke", "none");
  propList.insert("draw:fill", "none");

  if (style.pen)
  {
    writePen(propList, style.pen.value());
  }

  if (style.fill && !noFill)
  {
    writeFill(propList, style.fill.value());
  }

  if (style.shadow)
  {
    writeShadow(propList, style.shadow.value());
  }
}

void ZMFCollector::writePen(librevenge::RVNGPropertyList &propList, const Pen &pen)
{
  propList.insert("svg:stroke-color", pen.color.toString());
  if (!ZMF_ALMOST_ZERO(pen.width))
    propList.insert("svg:stroke-width", pen.width);
  propList.insert("draw:stroke", "solid"); // todo: other line types

  switch (pen.lineCapType)
  {
  default:
  case LineCapType::BUTT:
    propList.insert("svg:stroke-linecap", "butt");
    break;
  case LineCapType::ROUND:
    propList.insert("svg:stroke-linecap", "round");
    break;
  case LineCapType::FLAT:
    propList.insert("svg:stroke-linecap", "square");
    break;
  }

  switch (pen.lineJoinType)
  {
  default:
  case LineJoinType::BEVEL:
    propList.insert("svg:stroke-linejoin", "bevel");
    break;
  case LineJoinType::MITER:
    propList.insert("svg:stroke-linejoin", "miter");
    break;
  case LineJoinType::ROUND:
    propList.insert("svg:stroke-linejoin", "round");
    break;
  }

  if (m_style.transparency)
  {
    propList.insert("svg:stroke-opacity", m_style.transparency.value().opacity(), librevenge::RVNG_PERCENT);
  }
}

void ZMFCollector::writeFill(librevenge::RVNGPropertyList &propList, const Fill &fill)
{
  FillWriter fillWriter(propList, m_style.transparency);
  boost::apply_visitor(fillWriter, fill);

  propList.insert("svg:fill-rule", "evenodd");
}

void ZMFCollector::writeShadow(librevenge::RVNGPropertyList &propList, const Shadow &shadow)
{
  propList.insert("draw:shadow", "visible");
  propList.insert("draw:shadow-color", shadow.color.toString());
  propList.insert("draw:shadow-opacity", shadow.opacity, librevenge::RVNG_PERCENT);
  propList.insert("draw:shadow-offset-x", shadow.offset.x);
  propList.insert("draw:shadow-offset-y", shadow.offset.y);
}

}

/* vim:set shiftwidth=2 softtabstop=2 expandtab: */
