/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * This file is a part of the libzmf project.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#ifndef ZMFTYPES_H_INCLUDED
#define ZMFTYPES_H_INCLUDED

#include <vector>

#include <boost/optional.hpp>
#include <boost/variant.hpp>

#include "libzmf_utils.h"

namespace libzmf
{

struct Point
{
  double x;
  double y;

  Point()
    : x(0.0), y(0.0)
  { }

  Point(double xVal, double yVal)
    : x(xVal), y(yVal)
  { }
};

bool operator==(const Point &lhs, const Point &rhs);
bool operator!=(const Point &lhs, const Point &rhs);

enum class CurveType
{
  LINE,
  BEZIER_CURVE
};

struct Curve
{
  std::vector<Point> points;
  std::vector<CurveType> sectionTypes;
  bool closed;

  Curve()
    : points(), sectionTypes(), closed(false)
  {  }
};

struct Color
{
  uint8_t red;
  uint8_t green;
  uint8_t blue;

  Color()
    : red(0), green(0), blue(0)
  {  }

  Color(uint8_t r, uint8_t g, uint8_t b)
    : red(r), green(g), blue(b)
  {  }

  librevenge::RVNGString toString() const;
};

enum class LineCapType
{
  BUTT,
  FLAT,
  ROUND,
  POINTED
};

enum class LineJoinType
{
  MITER,
  ROUND,
  BEVEL
};

struct Pen
{
  Color color;
  double width;
  LineCapType lineCapType;
  LineJoinType lineJoinType;

  Pen()
    : color(), width(0.0),
      lineCapType(LineCapType::BUTT), lineJoinType(LineJoinType::MITER)
  {  }
};

struct GradientStop
{
  Color color;
  double offset;

  GradientStop()
    : color(), offset(0.0)
  { }
};

enum class GradientType
{
  LINEAR,
  RADIAL,
  CONICAL,
  CROSS,
  RECTANGULAR,
  FLEXIBLE
};

struct Gradient
{
  GradientType type;
  std::vector<GradientStop> stops;
  double angle;
  Point center;

  Gradient()
    : type(), stops(), angle(0.0), center(0.5, 0.5)
  { }
};

typedef boost::variant<Color, Gradient> Fill;

struct Transparency
{
  Color color;

  double opacity() const
  {
    return 1.0 - color.red / 255.0;
  }

  Transparency()
    : color()
  { }
};

struct Shadow
{
  Point offset;
  double angle;
  double opacity;
  Color color;

  Shadow()
    : offset(), angle(0.0), opacity(1.0), color()
  { }
};

struct Style
{
  boost::optional<Pen> pen;
  boost::optional<Fill> fill;
  boost::optional<Transparency> transparency;
  boost::optional<Shadow> shadow;

  Style()
    : pen(), fill(), transparency(), shadow()
  { }
};

struct ZMFPageSettings
{
  double width;
  double height;
  double leftOffset;
  double topOffset;
  Color color;

  ZMFPageSettings()
    : width(0.0), height(0.0), leftOffset(0.0), topOffset(0.0), color()
  { }
  ZMFPageSettings(double w, double h, double left, double top, Color c)
    : width(w), height(h), leftOffset(left), topOffset(top), color(c)
  { }
};

}

#endif // ZMFTYPES_H_INCLUDED

/* vim:set shiftwidth=2 softtabstop=2 expandtab: */
